// Copyright 1997 Ubilab, Union Bank of Switzerland. All rights reserved.
// This is part of the Geo system developed by Ubilab's DistrOOSE group.
// Check out: http://www.ubs.com/ubilab/Projects/se/se.html. GLAB-REH-97.

package Geo.Framework.AnyImpl;

import java.util.Vector;
import java.util.Enumeration;

import Geo.Framework.Any.*;



public class AnyArrayImpl extends AnyImpl implements AnyArray {

	// instance variables
	
	protected Vector contents;


	// constructors
	
	public AnyArrayImpl() {
		contents = new Vector();
	}
	
	public AnyArrayImpl(int initialCapacity) {
		contents = new Vector(initialCapacity);
	}
	
	protected AnyArrayImpl(Enumeration anEnumeration) {
		this();
		Object next;
		while (anEnumeration.hasMoreElements()) {
			next = anEnumeration.nextElement();
			if (next instanceof Any || next == null) 
				append((Any) next);
		}
	}
	
	protected AnyArrayImpl(Vector aVector) {
		contents = (Vector) aVector.clone();
	}
	
	
	// accessing

	public int size() {
		return contents.size();
	}

	
	// accessing elements
	
	public Any at(int pos) {
		return (Any) contents.elementAt(pos);
	}
	
	
	// manipulating elements
	
	public int append(Any newElement) {
		if (newElement == null) newElement = AnySap.nullValue();
		contents.addElement(newElement);
		return size()-1;
	}

	public void appendArray(AnyArray newElements) {
		for (int i=0, s=newElements.size(); i<s; i++) 
			append(newElements.at(i)); 
	}
	
	public Any atPut(int pos, Any newElement) {
		// size of AnyArray is dynamically adjusted to accomodate request if needed
		if ( pos >= size() ) this.expand(pos+1);
		if (newElement == null) newElement = AnySap.nullValue();
		contents.setElementAt(newElement, pos); 
		return newElement;
	}

	public void insert(int pos, Any newElement) {
		if (newElement == null) newElement = AnySap.nullValue();
		contents.insertElementAt(newElement, pos);
	}

	public Any remove(int pos) {
		Any removedElement;
		if (pos < 0 || pos >= size()) return null;
		removedElement = at(pos);
		contents.removeElementAt(pos);
		return removedElement;
	}
	
	public void remove(Any element) {
		for (int i=0, s = size(); i<s; i++) {
			if ( at(i) == element ) {
				remove(i);
				s--;
			}
		}
	}
	
	public void removeAll() {
		contents.removeAllElements();
	}
	
	
	// finding elements

	public AnyArray findAll(Any element) {
		AnyArray matches = new AnyArrayImpl();
		int index = findFirstPosition(element);
		while ( index != -1 ) {
			matches.append(at(index));
			index = contents.indexOf(element, index+1);
		}
		return matches;
	}
	
	public AnyArray findAllWithSlot(AnyString key, Any anAny) {  
		Any element;
		AnyArray matches = new AnyArrayImpl();
		for (int i=0, s=size(); i<s; i++) {
			element = this.at(i);
			if ( element instanceof AnyDictionary && ((AnyDictionary)element).at(key).isEqual(anAny) 
				|| element instanceof AnyFrame && ((AnyFrame)element).at(key).isEqual(anAny) )
						matches.append(element); 
		}		
		return matches;
	}
	
	public AnyArray findAllWithSlot(String key, Any anAny) {  
		return findAllWithSlot(AnySap.stringValue(key), anAny);
	}
	
	public Any findFirst(Any element) {
		int index = findFirstPosition(element);
		return index == -1 ? null : at(index); 
	}
	
	public int findFirstPosition(Any element) {
		// returns -1 if not found
		return contents.indexOf(element);
	}
	
	public Any findFirstWithSlot(AnyString key, Any anAny) {		
		Any element;											
		for (int i=0, s=size(); i<s; i++) {				
			element = at(i);
			if ( element instanceof AnyDictionary && ((AnyDictionary)element).at(key).isEqual(anAny) 
				|| element instanceof AnyFrame && ((AnyFrame)element).at(key).isEqual(anAny) )
						return element; 			
		}
		return null;
	}
	
	public Any findFirstWithSlot(String key, Any anAny) {	
		return findFirstWithSlot(AnySap.stringValue(key), anAny);
	}	


	// comparing
	
	public boolean isEqual(Any anAny) {
		if ( anAny == null || !(anAny instanceof AnyArray) )
			return false;
		AnyArray anAnyArray = anAny.asAnyArray();
		if ( size() != anAnyArray.size() )
			return false;
		for (int i=0, s=size(); i<s; i++) {
			if ( at(i).notEqual(anAnyArray.at(i)) ) return false;		
		}
		return true;
	}

	public int hashCode() {
		int hashCode = 0;
		for (int i=0, s=size(); i<s; i++) 
			hashCode = hashCode + at(i).hashCode();
		return hashCode;
	}

	
	// copying

	public Any shallowCopy() {
		AnyArrayImpl newAnyArray = (AnyArrayImpl) super.shallowCopy();
		newAnyArray.contents = (Vector) contents.clone();	
		return (Any) newAnyArray;
	}
		
	public Any deepCopy() {
		AnyArrayImpl newAnyArray = (AnyArrayImpl) super.deepCopy();
		newAnyArray.contents = new Vector();
		for (int i=0, s=size(); i<s; i++) 
			newAnyArray.append(at(i).deepCopy());
		return (Any) newAnyArray;	
	}	
	
	
	// testing
	
	public boolean isSimple() {
		return false;
	}


	// printing
	
	public String toString() {
		return "anAnyArray";
	}
	
		
	// misc

	public void addAllTo(AnyArray anAnyArray) {
		super.addAllTo(anAnyArray);
		for (int i=0, s=size(); i<s; i++) 
			at(i).addAllTo(anAnyArray);
	}

	public void expand(int newSize) {
		int oldSize = size();
		if (newSize > oldSize) {
			contents.setSize(newSize);
			for (int i=oldSize, s=newSize-1; i<s; i++)
				contents.setElementAt(AnySap.nullValue(), i);
		}
	}

	public void sort() {
		quicksortWithKey(0, size()-1, null);
	}
	
	public void sort(AnyString key) {
		quicksortWithKey(0, size()-1, key);
	}
	
	public void sort(String key) {
		sort(AnySap.stringValue(key));
	}


	// private

	private void quicksortWithKey(int left, int right, AnyString slotKey ) {
		int i, last;
		boolean doSwap;
		if (left >= right) return;		  	
		swap(left, (left+right) / 2);
		last = left;
		for (i = left+1; i <= right; i++) {
			if ( slotKey == null ) 
				doSwap = compare( at(i), at(left) ) < 0;
			else
				doSwap = compare( at(i), at(left), slotKey ) < 0;
			if (doSwap) swap(++last, i);
		}	
		swap(left, last);
		quicksortWithKey(left, last-1, slotKey);
		quicksortWithKey(last+1, right, slotKey);
	}

	private int compare(Any a, Any b) {
		if ( a instanceof AnyString && b instanceof AnyString) 
			return a.toString().compareTo(b.toString());
		if (a instanceof AnyInteger && b instanceof AnyInteger)
			return a.asInteger() - b.asInteger();
		if (a instanceof AnyDouble && b instanceof AnyDouble) 
			return (int) (a.asDouble() - b.asDouble());
		if (a instanceof AnyIndexCacheSlot && b instanceof AnyIndexCacheSlot)
			return ((AnyIndexCacheSlot) a).slotName().asString().compareTo(((AnyIndexCacheSlot) b).slotName().asString());
		return 0;
	}
	
	private int compare(Any a, Any b, AnyString slotKey) {
		if ( a instanceof AnyDictionary && b instanceof AnyDictionary ) 
			return compare(a.asAnyDictionary().at(slotKey), b.asAnyDictionary().at(slotKey));
		if ( a instanceof AnyFrame && b instanceof AnyFrame ) 
			return compare(a.asAnyFrame().at(slotKey), b.asAnyFrame().at(slotKey));
		return 0;
	}
	
	private void swap(int i, int j) {
		Any tmp = at(i);
		atPut(i, at(j));
		atPut(j, tmp);
	}	

}
