// Copyright 1997 Ubilab, Union Bank of Switzerland. All rights reserved.
// This is part of the Geo system developed by Ubilab's DistrOOSE group.
// Check out: http://www.ubs.com/ubilab/Projects/se/se.html. GLAB-REH-97.

package Geo.Framework.AnyImpl;

import java.util.Hashtable;
import Geo.Framework.Any.*;



public class AnyDictionaryImpl extends AnyImpl implements AnyDictionary {

	// instance variables
	
	protected Hashtable contents;


	// constructors
	
	public AnyDictionaryImpl() {
		contents = new Hashtable();
	}


	// accessing

	public AnyArray keys() {
		return new AnyArrayImpl(contents.keys());
	}

	public AnyArray values() {
		return new AnyArrayImpl(contents.elements());
	}
	
	public int size() {
		return contents.size();
	}	

	public int fullSize() {
		int fullSize = 0;
		AnyArray keys = keys();
		for (int i=0,s=keys.size(); i<s; i++) 
			fullSize += sizeAt(keys.at(i).asAnyString());
		return fullSize;
	}	
	
	
	// accessing slots

	public Any at(AnyString key) {
		return at(key,0); 
	}
	
	public Any at(String key) {
		return at(AnySap.stringValue(key));
	}
	
	public Any at(AnyString key, int pos) {
		Any value = slotAt(key);
		if (value==null) return null;
		if (value instanceof AnyArray) 
			return value.asAnyArray().at(pos);
		else {
			if (pos==0) return value;
			else throw new IndexOutOfBoundsException();
		}
	}

	public Any at(String key, int pos) {
		return at(AnySap.stringValue(key), pos);
	}	
	
	public Any slotAt(AnyString key) {
		return (Any) contents.get(key);
	}
	
	public Any slotAt(String key) {
		return slotAt(AnySap.stringValue(key));
	}
	
	public AnyArray arrayAt(AnyString key) {
		Any value = slotAt(key);
		if (value == null) return null;
		if (value instanceof AnyArray)
			return value.asAnyArray();
		else 
			return inAnyArray(value);
	}
	
	public AnyArray arrayAt(String key) {
		return arrayAt(AnySap.stringValue(key));
	}

	public int sizeAt(AnyString key) {
		Any value = slotAt(key);
		if (value == null) return -1;
		if (value instanceof AnyArray)
			return value.asAnyArray().size();
		else
			return 1;
	}
	
	public int sizeAt(String key) {
		return sizeAt(AnySap.stringValue(key));
	}	
	
	
	// manipulating slots

	public void append(AnyString key, Any newElement) {
		if (newElement == null)
			throw new AnyNullPointerException(this);
		Any value = slotAt(key);
		if (value == null) 
			atPut(key, newElement);
		else {
			if (value instanceof AnyArray) 
				value.asAnyArray().append(newElement);
			else {
				AnyArray tmp = inAnyArray(value);
				tmp.append(newElement);
				setSlot(key, tmp);
			}
		}
	}
	
	public void append(String key, Any newElement) {
		append(AnySap.stringValue(key), newElement);
	}

	public void appendArray(AnyString key, AnyArray newElements) {
		for (int i=0, s=newElements.size(); i<s; i++) {
			if (newElements.at(i) == null) 
				throw new AnyNullPointerException(this);
		}
		Any value = slotAt(key);
		// make sure that value is an array first
		if (value == null) 	
			setSlot(key, new AnyArrayImpl());
		else {
			if ( !(value instanceof AnyArray) )
				setSlot(key, inAnyArray(value));
		}
		slotAt(key).asAnyArray().appendArray(newElements); 	
	}
	
	public void appendArray(String key, AnyArray newElements) {
		appendArray(AnySap.stringValue(key), newElements);
	}

	public void atPut(AnyString key, Any newElement) {
		atPut(key, 0, newElement);
	}
	
	public void atPut(String key, Any newElement) {
		atPut(AnySap.stringValue(key), newElement);
	}
	
	public void atPut(AnyString key, int pos, Any newElement) { 
		// value is dynamically expanded to accomodate request
		if (newElement == null) throw new AnyNullPointerException(this);
		Any value = slotAt(key);
		if ( value == null || !(value instanceof AnyArray) ) {
			if (pos == 0) {
				if (newElement instanceof AnyArray) 
					setSlot(key, inAnyArray(newElement));
				else
					setSlot(key, newElement); }
			else {
				AnyArray tmp = inAnyArray(value);
				tmp.atPut(pos, newElement);
				setSlot(key, tmp); }} 
		else  // it's an AnyArray
			value.asAnyArray().atPut(pos, newElement);
	}

	public void atPut(String key, int pos, Any newElement) { 
		atPut(AnySap.stringValue(key), pos, newElement);
	}
	
	public void setSlot(AnyString key, Any value) {
		if (value == null) 
			throw new AnyNullPointerException(this);
		contents.put(key, value);
	}
	
	public void setSlot(String key, Any value) {
		setSlot( AnySap.stringValue(key), value);
	}	

	public Any remove(AnyString key) {
		return remove(key, 0);
	}

	public Any remove(String key) {
		return remove(AnySap.stringValue(key));
	}	
	
	public Any remove(AnyString key, int pos) {
		Any value = slotAt(key);
		if (value == null) return null;
		if (value instanceof AnyArray) {
			if ( pos == 0 && sizeAt(key) == 1 ) 	
				return removeSlot(key).at(0);
			else
				return value.asAnyArray().remove(pos); }
		else {
			if (pos == 0) return removeSlot(key).at(0);
			else throw new IndexOutOfBoundsException();
		}
	}

	public Any remove(String key, int pos) {
		return remove(AnySap.stringValue(key), pos);
	}
	
	public AnyArray removeSlot(AnyString key) {
		Any removedAny = (Any) contents.remove(key);
		if ( removedAny instanceof AnyArray )
			return removedAny.asAnyArray();
		if ( removedAny == null ) 
			return null;
		return inAnyArray(removedAny);
	}
	
	public AnyArray removeSlot(String key) {
		return removeSlot(AnySap.stringValue(key));
	}


	// comparing

	public boolean isEqual(Any anAny) {
		if ( anAny == null || !(anAny instanceof AnyDictionary) )
			return false;
		AnyDictionary anAnyDictionary = anAny.asAnyDictionary();
		if ( size() != anAnyDictionary.size() )
			return false;
		return keys().isEqual(anAnyDictionary.keys()) && values().isEqual(anAnyDictionary.values());
	}
	
	public int hashCode() {
		return keys().hashCode() + values().hashCode(); 
	}	
	
	
	// copying

	public Any shallowCopy() {
		AnyDictionaryImpl newAnyDictionary = (AnyDictionaryImpl) super.shallowCopy();
		newAnyDictionary.contents = new Hashtable();
		AnyArray keys = keys();
		AnyArray values = values();
		for (int i=0, s=keys.size(); i<s; i++) {
			Any clonedValue;
			Any value = values.at(i);
			if (value instanceof AnyArray)
				clonedValue = value.shallowCopy();
			else
				clonedValue = value;
			newAnyDictionary.setSlot(keys.at(i).asAnyString(), clonedValue);
		}
		return newAnyDictionary;
	}		
	
	public Any deepCopy() {
		AnyDictionaryImpl newAnyDictionary = (AnyDictionaryImpl) super.deepCopy();
		newAnyDictionary.contents = new Hashtable();
		AnyArray keys = keys();
		AnyArray values = values();
		for (int i=0, s=keys.size(); i<s; i++) 
			newAnyDictionary.setSlot(keys.at(i).asAnyString(), values.at(i).deepCopy());
		return newAnyDictionary;
	}
	
	
	// testing

	public boolean isSimple() {
		return false;
	}

	public boolean hasSlot(AnyString key) {
		return at(key) != null;
	}
	
	public boolean hasSlot(String key) {
		return hasSlot(AnySap.stringValue(key));
	}


	// printing

	public String toString() {
		return "anAnyDictionary";
	}

	
	// misc
	
	public void addAllTo(AnyArray anAnyArray) {
		AnyArray values = values();
		super.addAllTo(anAnyArray);
		for (int i=0, s=values.size(); i<s; i++) {
			Any value = values.at(i); 
			if ( value instanceof AnyArray ) {
				AnyArray valueArray = value.asAnyArray();
				for (int j=0, t=valueArray.size(); j<t; j++) 
					valueArray.at(j).addAllTo(anAnyArray); } 
			else
				value.addAllTo(anAnyArray);
		}
	}
	
	
	// private
	
	private AnyArray inAnyArray(Any anAny) {
		// return new array whose only element is anAny
		AnyArray anyArray = new AnyArrayImpl();
		anyArray.append(anAny);
		return anyArray;
	}
	
}
