// Copyright 1997 Ubilab, Union Bank of Switzerland. All rights reserved.
// This is part of the Geo system developed by Ubilab's DistrOOSE group.
// Check out: http://www.ubs.com/ubilab/Projects/se/se.html. GLAB-REH-97.

package Geo.Framework.AnyImpl;

import java.util.Vector;
import Geo.Framework.Any.*;



public class AnyFrameImpl extends AnyImpl implements AnyFrame {

	// instance variables
	
	protected AnyFrameDescriptor frameDescriptor;
	protected Vector contents;
	
	
	// constructors

	protected AnyFrameImpl() {
		contents = new Vector();
	}
	
	public AnyFrameImpl(AnyContext context, AnyString typeName) {
		this();
		this.frameDescriptor = context.getSymbolTable().at(typeName);
		if (getFrameDescriptor() == null) 
			throw new AnyTypeException(typeName.asString(), context); 
		initialize();
	}
	
	public AnyFrameImpl(AnyContext context, AnyString typeName, AnyArray slotValues) {
		this();
		this.frameDescriptor = context.getSymbolTable().at(typeName);
		if (getFrameDescriptor() == null) 
			throw new AnyTypeException(typeName.asString(), context); 
		for (int i=0, s=slotValues.size(); i<s; i++) {
			Any newValue = slotValues.at(i);
			if (newValue == null) newValue = AnySap.nullValue();
			contents.addElement(newValue);
		}
		initialize();
	}
	
	public AnyFrameImpl(AnyContext context, String typeName) {
		this(context, AnySap.stringValue(typeName));
	}
	
	public AnyFrameImpl(AnyContext context, String typeName, AnyArray slotValues) {
		this(context, AnySap.stringValue(typeName), slotValues);
	}
	
	public AnyFrameImpl(AnyFrameDescriptor descriptor) {
		this();
		this.frameDescriptor = descriptor;
		initialize();
	}
	
	
	// accessing
	
	public AnyContext getContext() {
		return getFrameDescriptor().getContext();
	}	
	
	public AnyFrameDescriptor getFrameDescriptor() {
		return frameDescriptor;
	}
	
	public AnyString getTypeName() {
		return getFrameDescriptor().getName(); 
	}

	public int size() {
		return contents.size();
	}	

	public int fullSize() {
		int fullSize = 0;
		AnyArray slotNames = slotNames();
		for (int i=0, s=slotNames.size(); i<s; i++) 
			fullSize += sizeAt(slotNames.at(i).asAnyString());
		return fullSize;
	}

	public AnyArray slotNames() {
		return getFrameDescriptor().allSlotNames();
	}
	
	public AnyArray slotValues() {
		return new AnyArrayImpl(contents);
	}
	
	
	// accessing slots
	
	public Any at(AnyString slotName) {
		return at(slotName, 0);
	}
	
	public Any at(String slotName) {
		return at(AnySap.stringValue(slotName));
	}	
	
	public Any at(AnyString slotName, int pos) {
		Any value = slotAt(slotName);
		if ( value instanceof AnyArray )
			return value.asAnyArray().at(pos);
		else{		
			if (pos==0) return value;
			else throw new IndexOutOfBoundsException();
		}
	}

	public Any at(String slotName, int pos) {
		return at(AnySap.stringValue(slotName), pos);
	}	

	public Any slotAt(AnyString slotName) {
		int pos = slotPositionFor(slotName);
		return (Any) contents.elementAt(pos);
	}
	
	public Any slotAt(String slotName) {
		return slotAt(AnySap.stringValue(slotName));
	}

	public AnyArray arrayAt(AnyString slotName) {
		Any value = slotAt(slotName);
		if ( value instanceof AnyArray ) 
			return value.asAnyArray();
		else
			return inAnyArray(value);
	}
	
	public AnyArray arrayAt(String slotName) {
		return arrayAt(AnySap.stringValue(slotName));
	}

	public int sizeAt(AnyString slotName) {
		Any value = slotAt(slotName);
		if (value instanceof AnyArray) 
			return value.asAnyArray().size();
		else
			if (value.isEqual(null)) return 0;
			else return 1;
	}
	
	public int sizeAt(String slotName) {
		return sizeAt(AnySap.stringValue(slotName));
	}

	
	// manipulating slots
	
	public void append(AnyString slotName, Any newElement) {
		Any value = slotAt(slotName);
		if (value.isEqual(null)) 
			atPut(slotName, newElement);
		else {
			if (value instanceof AnyArray) 
				value.asAnyArray().append(newElement);
			else {
				AnyArray tmp = inAnyArray(value); 
				tmp.append(newElement);
				setSlot(slotName, tmp); 
			}
		}
	}
	
	public void append(String slotName, Any newElement) {
		append(AnySap.stringValue(slotName), newElement);
	}	
	
	public void appendArray(AnyString slotName, AnyArray newElements) {
		for (int i=0, s=newElements.size(); i<s; i++) {
			if (newElements.at(i) == null)
				newElements.atPut(i, AnySap.nullValue());
		}
		Any value = slotAt(slotName);
		if (value.isEqual(null)) 
			setSlot(slotName, newElements);
		else {
			if ( !(value instanceof AnyArray) )
				setSlot(slotName, inAnyArray(value));
			slotAt(slotName).asAnyArray().appendArray(newElements);
		}
	}
	
	public void appendArray(String slotName, AnyArray newElements) {
		appendArray(AnySap.stringValue(slotName), newElements);
	}

	public void atPut(AnyString slotName, Any newElement) {
		atPut(slotName, 0, newElement);
	}
	
	public void atPut(String slotName, Any newElement) {
		atPut(AnySap.stringValue(slotName), newElement);
	}
	
	public void atPut(AnyString slotName, int pos, Any newElement) {
		Any value = slotAt(slotName);
		if (value instanceof AnyArray)
			value.asAnyArray().atPut(pos, newElement);
		else {
			if (pos==0) {
				if (newElement instanceof AnyArray) 
					setSlot(slotName, inAnyArray(newElement));
				else
					setSlot(slotName, newElement); }
			else { 
				if (value.isEqual(null))
					setSlot(slotName, newElement);
				else {
					AnyArray newAnyArray = inAnyArray(value);
					newAnyArray.atPut(pos, newElement);
					setSlot(slotName, newAnyArray); }
			}	
		}
	}

	public void atPut(String slotName, int pos, Any newElement) {
		atPut(AnySap.stringValue(slotName), pos, newElement);
	}

	public void setSlot(AnyString slotName, Any value) {
		if (value==null) value = AnySap.nullValue();
		int pos = slotPositionFor(slotName);
		contents.setElementAt(value, pos);
	}
	
	public void setSlot(String slotName, Any value) {
		setSlot(AnySap.stringValue(slotName), value);
	}

	public Any remove(AnyString slotName) {
		return remove(slotName, 0);
	}
	
	public Any remove(String slotName) {
		return remove(AnySap.stringValue(slotName));
	}
	
	public Any remove(AnyString slotName, int pos) {
		Any value = slotAt(slotName);
		if (value instanceof AnyArray) {
			if ( pos==0 && value.asAnyArray().size()==1 ) {
				Any removedElement = at(slotName);
				setSlot(slotName, null);
				return removedElement; }
			else
				return value.asAnyArray().remove(pos); }
		else {
			if (pos==0) {
				setSlot(slotName, null);
				return value; }
			else throw new IndexOutOfBoundsException();
		}
	}

	public Any remove(String slotName, int pos) {
		return remove(AnySap.stringValue(slotName), pos);
	}


	// comparing

	public boolean isEqual(Any anAny) {
		if ( anAny == null || !(anAny instanceof AnyFrame) )
			return false;
		AnyFrame anAnyFrame = anAny.asAnyFrame(); 
		return getFrameDescriptor().isEqual(anAnyFrame.getFrameDescriptor())	
			&& slotValues().isEqual(anAnyFrame.slotValues());
	}

	public int hashCode() {
		return slotValues().hashCode() + getFrameDescriptor().hashCode();
	}
	
	
	// copying

	public Any shallowCopy() {
		AnyFrameImpl newAnyFrame = (AnyFrameImpl) super.shallowCopy();
		newAnyFrame.frameDescriptor = frameDescriptor;
		newAnyFrame.contents = new Vector();
		newAnyFrame.contents.setSize(size());
		AnyArray allSlotNames = slotNames();
		for (int i=0, s=size(); i<s; i++) {	 
			Any clonedValue;
			Any value = (Any) contents.elementAt(i);
			if (value instanceof AnyArray)
				clonedValue = value.shallowCopy();
			else 
				clonedValue = value;
			newAnyFrame.setSlot(allSlotNames.at(i).asAnyString(), clonedValue); 
		}
		return (Any) newAnyFrame;	
	}
	
	public Any deepCopy() {
		AnyFrameImpl newAnyFrame = (AnyFrameImpl) super.deepCopy();
		newAnyFrame.frameDescriptor = frameDescriptor;
		newAnyFrame.contents = new Vector();
		newAnyFrame.contents.setSize(size());
		AnyArray allSlotNames = slotNames();
		for (int i=0, s=size(); i<s; i++) {
			Any clonedValue = ((Any) contents.elementAt(i)).deepCopy();
			newAnyFrame.setSlot(allSlotNames.at(i).asAnyString(), clonedValue); }
		return (Any) newAnyFrame;	
	}	
	
	// testing

	public boolean isSimple() {
		return false;
	}

	public boolean hasSlot(AnyString slotName) {
		int pos = getFrameDescriptor().offset(slotName);
		if ( pos == -1 )
			return false;
		else
			return true;
		}
		
	public boolean hasSlot(String slotName) {
		return hasSlot(AnySap.stringValue(slotName));
	}	

	public boolean isA(AnyString descriptorName) {
		return getTypeName().isEqual(descriptorName);
	}
	
	public boolean isA(String descriptorName) {
		return isA(AnySap.stringValue(descriptorName));
	}

	public boolean isKindOf(AnyString descriptorName) {
		AnyFrameDescriptor currentDescriptor = getFrameDescriptor();
		while ( !(currentDescriptor.getName().isEqual(descriptorName)) ) {
			currentDescriptor = currentDescriptor.getSuperFrameDescriptor();
			if (currentDescriptor == null) return false;
		}
		return true;
	}
	
	public boolean isKindOf(String descriptorName) {
		return isKindOf(AnySap.stringValue(descriptorName));
	}
	
	
	// printing
	
	public String toString() {
		return "anAnyFrame";
	}
	
	
	// misc
	
	public void addAllTo(AnyArray anAnyArray) {
		anAnyArray.append(this);
		for (int i=0, s=size(); i<s; i++) {
			if ( contents.elementAt(i) != null ) {
				if ( contents.elementAt(i) instanceof AnyArray ) {
					AnyArray elements = (AnyArray) contents.elementAt(i);
					for (int j=0, t=elements.size(); j<t; j++)
						elements.at(j).addAllTo(anAnyArray); }
				else
					((Any) contents.elementAt(i)).addAllTo(anAnyArray);
			}
		}
	}
	
	
	// private / protected
	
	private AnyArray inAnyArray(Any anAny) {
		AnyArray newAnyArray = new AnyArrayImpl();
		if ( anAny.notEqual(null) ) 
			newAnyArray.append(anAny);
		return newAnyArray;
	}

	private int slotPositionFor(AnyString slotName) {
		int pos = getFrameDescriptor().offset(slotName);
		if ( pos == -1 )
			throw new AnyFrameSlotException(slotName.asString(), getFrameDescriptor());
		else
			return pos;
	}

	protected void initialize() {
		contents.setSize(getFrameDescriptor().totalNumberOfSlots());
		for (int i=0, s=contents.size(); i<s; i++) {
			if (contents.elementAt(i) == null)
				contents.setElementAt(AnySap.nullValue(), i);
		}
	}

}
