// Copyright 1997 Ubilab, Union Bank of Switzerland. All rights reserved.
// This is part of the Geo system developed by Ubilab's DistrOOSE group.
// Check out: http://www.ubs.com/ubilab/Projects/se/se.html. GLAB-REH-97.

package Geo.Framework.AnyImpl;

import Geo.Framework.Any.*;



public abstract class InfoServiceDefImpl implements InfoService {

	// instance variables
	
	protected  AnySoup soup;

	
	// constructors
	
	protected InfoServiceDefImpl(AnySoup soup) {
		this.soup = soup;
	}
	
	
	// exploring soup

	public AnyArray getByKind(String descriptorName) { 
		AnyArray result = new AnyArrayImpl();
		for (int i=0, s=soup.getData().size(); i<s; i++) {	
			AnyFrame frame = soup.getData().at(i).asAnyFrame();
			if ( frame.isKindOf(descriptorName) )
				result.append(frame);
		}
		return result;
	}
	
	public AnyArray getBySlotStartingWith(String descriptorName, String slotName, String prefix) {
		AnyArray frames = getByKind(descriptorName);
		AnyArray result = new AnyArrayImpl();
		for (int i=0, s=frames.size(); i<s; i++) {	
			AnyFrame frame = frames.at(i).asAnyFrame();
			Any slotValue = frame.at(slotName);
			if ( slotValue instanceof AnyString && slotValue.asString().startsWith(prefix) )
				result.append(frame); 	
		}
		return result;
	}
	
	public AnyArray getBySlotValue(String descriptorName, String slotName, Any value) {
		AnyArray frames = getByKind(descriptorName);
		AnyArray result = new AnyArrayImpl();
		for (int i=0, s=frames.size(); i<s; i++) {	
			AnyFrame frame = frames.at(i).asAnyFrame();
			if ( frame.at(slotName).isEqual(value) )
				result.append(frame);
		}
		return result;
	}
	
	public AnyFrame getFirstBySlotValue(String descriptorName, String slotName, Any value) {
		for (int i=0, s=soup.getData().size(); i<s; i++) {	
			AnyFrame frame = soup.getData().at(i).asAnyFrame();
			if ( frame.isKindOf(descriptorName) && frame.at(slotName).isEqual(value) )
				return frame;
		}
		return null;
	}		
	
	// manipulating soup

	public void cleanSlot(String descriptorName, String slotName) {
		setSlot(descriptorName, slotName, null);
	}	

	public void setSlot(String descriptorName, String slotName, Any newValue) {
		AnyArray frames = getByKind(descriptorName);
		for (int i=0, s=frames.size(); i<s; i++) 
			frames.at(i).asAnyFrame().setSlot(slotName, newValue);
	}	
	
	
	// queries
	
	public AnyArray evaluateQuery(String queryExpression) {
		return soup.evaluateQuery(queryExpression);
	}

	public void removeByQuery(String queryExpression) {
		soup.removeByQuery(queryExpression);
	}
	
}
